@@ -1,75 +1,23 @@
-static void invalidate_sv(SV *ref, void *to)
-{
-  SV *iobj;
-  assert(ref);
-  assert(SvROK(ref));
-  iobj = SvRV(ref);
-  assert(SvTYPE(iobj) == SVt_PVHV);
-  assert(SvOBJECT(iobj));
-  HvNAME(iobj) = to;
-  /*warn("Invalidate (%s=0x%x)", HvNAME(SvSTASH(iobj)), iobj); /**/
-}
-
-static void unwrap_obj(SV *sv, void **vpp, int *is_event)
-{
-    SV *origsv = sv;
-    assert(vpp);
-    *vpp=0;
-    if (sv && SvROK(sv)) {
-	sv = SvRV(sv);
-	assert(sv);
-	if (SvOBJECT(sv) && SvTYPE(sv) == SVt_PVHV) {
-	    /* also can try xhv_array? */
-	    *vpp = HvNAME(sv);
-	}
-    } else {
-	croak("Not a reference?");
-    }
-    if (!*vpp)
-	croak("Attempt to use destroyed object (RV=0x%x %s=0x%x)",
-	      origsv, HvNAME(SvSTASH(sv)), sv);
-    if (is_event) *is_event = ((pe_watcher*)*vpp)->vtbl->base.is_event;
-}
-
-static void decode_sv(SV *sv, pe_watcher **wap, pe_event **evp)
-{
-  void *vp;
-  int is_event;
-  unwrap_obj(sv, &vp, &is_event);
-  assert(vp);
-  if (!is_event) {
-    assert(wap);
-    *wap = (pe_watcher *) vp;
-    if (evp) *evp=0;
-  }
-  else {
-    pe_event *ev = (pe_event*) vp;
-    assert(wap || evp);
-    if (evp) *evp = ev;
-    if (wap) *wap = ev->up;
-  }
-}
-
-static SV *wrap_tiehash(void *ptr, HV *stash)
-{
-  SV *tied = (SV*) newHV();
-  HvNAME(tied) = ptr;
-  /* HvARRAY(tied) = 0x12341234; /**/
-  sv_magic(tied, Nullsv, 'P', 0, 0);
-  return sv_bless(newRV_noinc(tied), stash);
+static SV *wrap_tiehash(void *ptr, HV *stash) {
+    SV *tied = (SV*) newHV();
+    HvNAME(tied) = ptr;
+    /* HvARRAY(tied) = 0x12341234; /**/
+    sv_magic(tied, Nullsv, 'P', 0, 0);
+    return sv_bless(newRV_noinc(tied), stash);
 }
 
 static SV *watcher_2sv(pe_watcher *wa) {
+    assert(!EvDESTROYED(wa));
     if (!wa->mysv) {
-	wa->mysv = wrap_tiehash(wa, wa->stash);
+	wa->mysv = wrap_tiehash(wa, wa->vtbl->stash);
     }
-    return wa->mysv;
+    return SvREFCNT_inc(sv_2mortal(wa->mysv));
 }
 
 static SV *event_2sv(pe_event *ev) {
     if (!ev->mysv) {
 	STRLEN n_a;
-	ev->mysv = wrap_tiehash(ev, ev->up->stash);
+	ev->mysv = wrap_tiehash(ev, ev->vtbl->stash);
 	if (EvDEBUGx(ev) >= 4)
 	    warn("Event=0x%x '%s' wrapped with SV=0x%x",
 		 ev, SvPV(ev->up->desc, n_a), SvRV(ev->mysv)); /**/
@@ -77,10 +25,23 @@ static SV *event_2sv(pe_event *ev) {
     return ev->mysv;
 }
 
+static void* unwrap_obj(SV *sv) {
+    /* switch back to '~' magic? XXX */
+    SV *origsv = sv;
+    if (!sv || !SvROK(sv))
+	croak("Not a reference?");
+    sv = SvRV(sv);
+    if (SvOBJECT(sv) && SvTYPE(sv) == SVt_PVHV)
+	return (void*) HvNAME(sv);
+    croak("Attempt to use destroyed object (RV=0x%x %s=0x%x)",
+	  origsv, HvNAME(SvSTASH(sv)), sv);
+    return 0;
+}
+
 static void get_base_vtbl(SV *sv, void **vp, pe_base_vtbl **vt) {
-  assert(vp && vt);
-  unwrap_obj(sv, vp, 0);
-  *vt = **(pe_base_vtbl***)vp;
+    assert(vp && vt);
+    *vp = unwrap_obj(sv);
+    *vt = **(pe_base_vtbl***)vp;
 }
 
 /***************************************************************/
@@ -155,8 +116,3 @@ static int sv_2events_mask(SV *sv, int bits)
     return 0; /* NOTREACHED */
   }
 }
-
-static void boot_typemap()
-{
-    /* nuke? XXX */
-}
